#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 1A.soft - Notions de SQL
# 
# Premiers pas avec le langage [SQL](https://fr.wikipedia.org/wiki/Structured_Query_Language).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# Le langage SQL est utilisé pour manipuler des [bases de données](https://fr.wikipedia.org/wiki/Base_de_donn%C3%A9es). Pour faire simple, on utilise les bases de données pour accéder rapidement à une information dans des données qui font parfois plusieurs milliards de lignes.
# 
# * Le tableau dont on se sert est trop grand (comme trier 50000 lignes).
# * On souhaite faire des opérations sur deux feuilles Excel (associer les lignes de l'une avec celles de l'autre).
# 
# Lorsque le volume de données est important, il est impossible de les voir dans leur ensemble. On peut en voir soit une partie soit une aggrégation. Par exemple, la société qui gère les vélib a ouvert l'accès à ses données. Il est possible de télécharger aussi souvent qu'on veut (toutes les minutes par exemple) un état complet des vélos et places disponibles pour toutes les stations de Paris : c'est une table qui s'enrichit de 1300 lignes toutes les minutes.

# ## Récupérer les donnée

# In[2]:


from pyensae.datasource import download_data
download_data("td8_velib.zip", website = 'xd')


# On crée une base de données [sqlite3](https://www.sqlite.org/). On peut la consulter avec un outil tel que [SqliteSpy](https://www.yunqa.de/delphi/products/sqlitespy/index) sous Windows, [sqlite_bro](https://pypi.python.org/pypi/sqlite_bro) sur tous les OS.

# In[3]:


from pyensae.sql import import_flatfile_into_database
dbf = "td8_velib.db3"
import_flatfile_into_database(dbf, "td8_velib.txt")                   # 2 secondes 
import_flatfile_into_database(dbf, "stations.txt", table="stations")  # 2 minutes


# Vous devriez voir un fichier .db3.

# In[4]:


import os
[ _ for _ in os.listdir(".") if ".db3" in _]


#  ## Premières requêtes SQL
#  
#  Dans notre cas, on va faire cela depuis le notebook.

# In[5]:


get_ipython().run_line_magic('load_ext', 'pyensae')


# In[6]:


get_ipython().run_line_magic('SQL_connect', 'td8_velib.db3')


# On regarde les tables de la base de données.

# In[7]:


get_ipython().run_line_magic('SQL_tables', '')


# On regarde les colonnes de chaque table.

# In[8]:


get_ipython().run_line_magic('SQL_schema', 'stations')


# In[9]:


get_ipython().run_line_magic('SQL_schema', 'td8_velib')


# Et enfin on regarde les premières lignes.

# In[10]:


get_ipython().run_cell_magic('SQL', '', 'SELECT * FROM td8_velib LIMIT 10\n')


#  On sélectionne les données sur une plage horaire donnée.

# In[11]:


get_ipython().run_cell_magic('SQL', '', "SELECT * FROM td8_velib WHERE last_update >= '2013-09-13 10:00:00' AND last_update <= '2013-09-13 11:00:00'\n")


# Sélectionner certaines colonnes et ordonner les valeurs.

# In[12]:


get_ipython().run_cell_magic('SQL', '', "SELECT available_bike_stands, available_bikes FROM td8_velib \nWHERE last_update >= '2013-09-13 10:00:00' AND last_update <= '2013-09-13 11:00:00'\nORDER BY available_bike_stands DESC ;\n")


# Compter le nombre d'emplacements de chaque station.

# In[13]:


get_ipython().run_cell_magic('SQL', '', "SELECT last_update, available_bike_stands + available_bikes AS place, number FROM td8_velib \nWHERE last_update >= '2013-09-13 10:00:00' AND last_update <= '2013-09-13 11:00:00'\nORDER BY place DESC ;\n")


# Par défaut la commande [%%SQL](http://www.xavierdupre.fr/app/pyensae/helpsphinx/i_nb.html#sql) n'affiche que les dix premières lignes.

# In[14]:


get_ipython().run_cell_magic('SQL', '--help', '\n.\n')


# On affiche 5 lignes et on stocke le résultat dans un dataframe.

# In[15]:


get_ipython().run_cell_magic('SQL', '-n 5 --df=df', "SELECT last_update, available_bike_stands + available_bikes AS place, number FROM td8_velib \nWHERE last_update >= '2013-09-13 10:00:00' AND last_update <= '2013-09-13 11:00:00'\nORDER BY place DESC ;\n")


# In[16]:


df.tail()


# Maximum de vélos disponibles à une station.

# In[17]:


get_ipython().run_cell_magic('SQL', '', 'SELECT MAX(available_bike_stands) FROM td8_velib\n')


# Et le minimum.

# In[18]:


get_ipython().run_cell_magic('SQL', '', 'SELECT "min" AS label, MIN(available_bike_stands) FROM td8_velib \nUNION ALL \nSELECT "max" AS label, MAX(available_bike_stands) FROM td8_velib \n')


# Tous les numéros de stations de façon unique.

# In[19]:


get_ipython().run_cell_magic('SQL', '', 'SELECT DISTINCT number FROM td8_velib\n')


# Compter le nombre de stations (1230).

# In[20]:


get_ipython().run_cell_magic('SQL', '', 'SELECT COUNT(*) FROM (\n    SELECT DISTINCT number FROM td8_velib\n)\n')


# ## Exercice 1
# 
# * Déterminer le nombre de valeur distinctes pour la colonne ``last_update``.
# * Déterminer la première et dernière date.

# ## GROUP BY
# 
# L'instruction ``GROUP BY`` permet d'aggréger des valeurs (min, max, sum) sur un ensemble de ligne partageant le même ensemble de valeurs (ou clé).

# In[21]:


get_ipython().run_cell_magic('SQL', '--df=df', 'SELECT last_update, SUM(available_bikes) AS velo_disponible\nFROM td8_velib\nGROUP BY last_update\nORDER BY last_update\n')


# Le résultat est un tableau avec de petites valeurs au début et de grandes vers la fin. Cela est dû au processus de création de la base de données. Certaines stations sont hors service et la dernière arrivée ou le dernier départ remonte à plusieurs jours. A chaque fois qu'on récupère les données velib, on dispose pour chaque station de la dernière arrivée ou du dernier départ de vélo. Le champ *last_update* correspond à cette date. Il ne faudra considérer que les dates au-delà de ``2013-09-10 11:30:19``.

# In[22]:


df.tail()


# ## Exercice 1b
# 
# Que fait la requête suivante ? Que se passe-t-il si vous enlevez les symboles ``--`` (on *décommente* la condition ``WHERE``) ?

# In[23]:


get_ipython().run_cell_magic('SQL', '--df=df', 'SELECT last_update, SUM(available_bikes) AS velo_disponible, COUNT(DISTINCT number) AS stations\nFROM td8_velib\n--WHERE last_update >= "2013-09-10 11:30:19"\nGROUP BY last_update\nORDER BY last_update\n')


# et celle-ci ?

# In[24]:


get_ipython().run_cell_magic('SQL', '--df=df', 'SELECT last_update, \n       CASE  WHEN available_bikes>0 THEN 1 ELSE 0 END AS vide, \n       COUNT(*) AS nb \nFROM td8_velib\nWHERE last_update >= "2013-09-10 11:30:19"\nGROUP BY last_update, vide\nORDER BY last_update\n')


# ## Exerice 2
# 
# Pour chaque station, compter le nombre de plages horaires de cinq minutes où il n'y a aucun vélo disponible.

# In[25]:





# ## Exercice 3
# 
# Si on note $X(s)$ le nombre de plages horaires de cinq minutes où il n'y a aucun vélo disponible, construire le tableau suivant : $k \rightarrow card\{ s | X(s) = k \}$.

# In[26]:





# ## JOIN
# 
# L'instruction ``JOIN`` sert à associer des lignes d'une table avec les lignes d'une autre table à partir du moment où elles partagent une information commune.

# In[27]:


get_ipython().run_cell_magic('SQL', '', 'SELECT A.*, B.name  -- ajout du nom au bout de chaque ligne\nFROM td8_velib AS A\nJOIN stations AS B\nON A.number == B.number\n')


# On peut s'en servir pour calculer un ratio en associant les deux instructions ``GROUP BY`` et ``JOIN``. L'instruction suivante permet d'obtenir la distribution des vélos disponibles sur la période d'étude pour chaque station.

# In[28]:


get_ipython().run_cell_magic('SQL', '', 'SELECT A.*, 1.0 * A.available_bikes / B.nb_velo  AS distribution_temporelle\nFROM td8_velib AS A\nJOIN (\n    SELECT number, SUM(available_bikes) AS nb_velo\n    FROM td8_velib\n    WHERE last_update >= "2013-09-10 11:30:19"\n    GROUP BY number\n) AS B\nON A.number == B.number\nWHERE A.last_update >= "2013-09-10 11:30:19"\n')


# ## Exercice 4 : distribution horaire
# 
# Pour chaque station, déterminer la distribution du nombre de vélos disponibles pour chaque période horaire d'une journée (par station, il y aura donc 24 * 12 valeurs comprises entre 0 et 1). Le résultat que vous devriez obtenir est illustré par l'image qui suit.

# In[29]:


from pyquickhelper.helpgen import NbImage
NbImage("images/tb8_dis_hor.png")


# ## Exercice 5 : zones de travail
# 
# On souhaite déterminer si une station se situe plutôt dans une zone de travail ou plutôt dans une zone de résidence. On part de l'hypothèse que, dans une zone de travail, les gens arrivent en vélib et repartent en vélib. C'est sans doute le cas de la station 8003. Les vélos seront plutôt disponibles dans la journée. A l'opposé, dans une zone de résidence, les vélos seront disponibles plutôt la nuit. Comment faire à partir de la distribution des vélos disponibles construite à la question précédente ?
# 
# On considère que la plage diurne s'étend de 10h à 16h. Vous trouverez une illustration du résultat dans cet [article](http://www.xavierdupre.fr/blog/2013-09-26_nojs.html).
# 

# In[30]:





# ## Exercice 6 : lattitude, longitude
# 
# On repart de la requête précédente pour effectuer un JOIN avec la table *stations* pour récupérer les coordonnées (lat, long). Après un copier/coller dans Excel, on peut situer les zones de travail sur la région parisienne. 

# In[31]:





# In[32]:


get_ipython().run_line_magic('SQL_close', '')


# ## Sans %%SQL
# 
# La commande magique ``%%SQL`` s'appuie sur le module [sqlite3](https://docs.python.org/3/library/sqlite3.html). On peut faire sans.

# In[33]:


import sqlite3
conn = sqlite3.connect("td8_velib.db3")       # on ouvre une connexion sur la base de données
data = conn.execute("SELECT * FROM stations") # on exécute une requête SQL
for i, d in enumerate(data):                  # on affiche le résultat
    print(d)
    if i > 5:
        break
conn.close()


# In[34]: